import asyncio
import time
from nsqio.utils import retry_iterator
from nsqio.tcp.connection import create_connection
from nsqio.tcp.consts import MPUB, PUB, SUB, AUTH, DPUB, INIT, CONNECTED, CLOSED
from nsqio.utils import get_logger

logger = get_logger()


async def create_writer(
    host="127.0.0.1",
    port=4150,
    loop=None,
    queue=None,
    heartbeat_interval=30000,
    feature_negotiation=True,
    tls_v1=False,
    snappy=False,
    deflate=False,
    deflate_level=6,
    consumer=False,
    sample_rate=0,
):
    """"
    param: host: host addr with no protocol. 127.0.0.1 
    param: port: host port 
    param: queue: queue where all the msg been put from the nsq 
    param: heartbeat_interval: heartbeat interval with nsq, set -1 to disable nsq heartbeat check
    params: snappy: snappy compress
    params: deflate: deflate compress  can't set True both with snappy
    """
    # TODO: add parameters type and value validation
    loop = loop or asyncio.get_event_loop()
    queue = queue or asyncio.Queue(loop=loop)
    writer = Writer(
        host=host,
        port=port,
        queue=queue,
        heartbeat_interval=heartbeat_interval,
        feature_negotiation=feature_negotiation,
        tls_v1=tls_v1,
        snappy=snappy,
        deflate=deflate,
        deflate_level=deflate_level,
        sample_rate=sample_rate,
        consumer=consumer,
        loop=loop,
    )
    await writer.connect()
    return writer


class Writer:
    def __init__(
        self,
        host="127.0.0.1",
        port=4150,
        loop=None,
        queue=None,
        heartbeat_interval=30000,
        feature_negotiation=True,
        tls_v1=False,
        snappy=False,
        deflate=False,
        deflate_level=6,
        sample_rate=0,
        consumer=False,
        max_in_flight=42,
    ):
        # TODO: add parameters type and value validation
        self._config = {
            "deflate": deflate,
            "deflate_level": deflate_level,
            "sample_rate": sample_rate,
            "snappy": snappy,
            "tls_v1": tls_v1,
            "heartbeat_interval": heartbeat_interval,
            "feature_negotiation": feature_negotiation,
        }

        self._host = host
        self._port = port
        self._conn = None
        self._loop = loop
        self._queue = queue or asyncio.Queue(loop=self._loop)
        self._status = INIT
        self._on_rdy_changed_cb = None
        self._is_working = True
        self._auto_reconnect_task_closed = asyncio.Event(loop=self._loop)
        self._auto_reconnect_task = self._loop.create_task(self.auto_reconnect())
        self._exit_event = asyncio.Event(loop=self._loop)

    async def connect(self):
        logger.debug("writer init connect")
        try:
            self._conn = await create_connection(
                self._host, self._port, self._queue, loop=self._loop
            )

            self._conn._on_message = self._on_message
            await self._conn.identify(**self._config)
            self._status = CONNECTED
        except Exception as e:
            logger.error("connect failed! {}".format(e))
            try:
                if self._conn:
                    self._conn.close()
            except Exception as e:
                logger.info("conn close failed, maybe its closed {}".format(e))
            finally:
                self._status = CLOSED

    def _on_message(self, msg):
        # should not be coroutine
        # update connections rdy state
        self.rdy_state = int(self.rdy_state) - 1

        self._last_message = time.time()
        if self._on_rdy_changed_cb is not None:
            self._on_rdy_changed_cb(self.id)
        return msg

    @property
    def last_message(self):
        return self._last_message

    async def reconnect(self):
        logger.debug("writer reconnect")
        logger.debug(self._status)
        try:
            if self._conn:
                self._conn.close()
            self._status = CLOSED
        except Exception as tmp:
            logger.info("conn close failed,maybe its closed already or init")
            logger.exception(tmp)
        await self.connect()

    async def auto_reconnect(self):
        logger.debug("writer autoreconnect")
        timeout_generator = retry_iterator(init_delay=0.1, max_delay=10.0)
        try:
            while self._is_working:
                logger.debug("writer autoreconnect check loop")
                if not (self._status == CONNECTED or self._status == INIT):
                    logger.debug(f"writer close({self._status})detected,reconnect")
                    conn_id = self.id if self._conn else "init"
                    logger.info("reconnect writer{}".format(conn_id))
                    try:
                        await self.reconnect()
                    except ConnectionError:
                        logger.error(
                            "Can not connect to: {}:{} ".format(self._host, self._port)
                        )
                    else:
                        self._status = CONNECTED
                t = next(timeout_generator)
                await asyncio.sleep(t, loop=self._loop)
        except asyncio.CancelledError:
            logger.info("{} auto_reconnect cancelled".format(self))
        finally:
            self._auto_reconnect_task_closed.set()
            try:
                if self._conn and not self._conn.closed:
                    self._conn.close()
            except Exception as tmp:
                logger.info("conn close failed,maybe its closed already or init")
                logger.exception(tmp)

    async def execute(self, command, *args, data=None):
        if self._conn.closed:
            logger.debug(f"execute found conn closed, reconnect()")
            await self.reconnect()
        response = self._conn.execute(command, *args, data=data)
        return await response

    async def auth(self, secret):
        """

        :param secret:
        :return:
        """
        return await self.execute(AUTH, data=secret)

    async def sub(self, topic, channel):
        """

        :param topic:
        :param channel:
        :return:
        """
        self._is_subscribe = True

        return await self.execute(SUB, topic, channel)

    async def pub(self, topic, message):
        """

        :param topic:
        :param message:
        :return:
        """
        return await self.execute(PUB, topic, data=message)

    async def dpub(self, topic, delay_time, message):
        """

        :param topic:
        :param message:
        :param delay_time: delayed time in millisecond
        :return:
        """
        if not delay_time or delay_time is None:
            delay_time = 0
        return await self.execute(DPUB, topic, delay_time, data=message)

    async def mpub(self, topic, *messages):
        """

        :param topic:
        :param message:
        :param messages:
        :return:
        """
        msgs = list(messages)
        return await self.execute(MPUB, topic, data=msgs)

    @property
    def id(self):
        return self._conn.endpoint

    async def close(self, timeout=10):
        # time_in = time.time()
        self._is_working = False
        self._conn.close()
        self._status = CLOSED
        try:
            await self._conn.wait_for_closed()
        except asyncio.TimeoutError:
            logger.warning("closing current connection failed: timeout")

        self._auto_reconnect_task.cancel()
        try:
            await asyncio.wait_for(
                self._auto_reconnect_task_closed.wait(),
                timeout=timeout,
                loop=self._loop,
            )
        except asyncio.TimeoutError:
            logger.warning("cancel auto_reconnect_task failed: timeout")
        logger.info("auto_reconnect_task closed")

    # def close(self, timeout = 10):
    #     time_in = time.time()
    #     close_task = self._loop.create_task(self.cancel(timeout))
    #     timeout_generator = retry_iterator(init_delay=0.01, max_delay=1.0)
    #     while True:
    #         if close_task.cancelled():
    #             logger.info("writer closer cancelled..")
    #             return
    #         if close_task.done():
    #             logger.info("writer closer finished..")
    #             return
    #         now = time.time()
    #         if timeout > 0 and now - time_in > timeout:
    #             logger.warning("writer closer timeout")
    #             return
    #         t = next(timeout_generator)
    #         time.sleep(t)

    def __repr__(self):
        return "<Writer{}>".format(self._conn.__repr__())



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from nsqio.tcp.connection import TcpConnection

import asyncio
import random
import logging
import time

from functools import partial

from nsqio.http import NsqLookupd
from nsqio.tcp.reader_rdy import RdyControl
from nsqio.tcp.connection import create_connection
from nsqio.tcp.consts import SUB, RDY, CLS
from nsqio.utils import get_logger, get_host_and_port, retry_iterator, get_version


logger = get_logger()


async def create_reader(
    nsqd_tcp_addresses=None,
    loop=None,
    max_in_flight=42,
    lookupd_http_addresses=None,
    **kwargs,
):
    """"
    initial function to get consumer
    param: nsqd_tcp_addresses: tcp addrs with no protocol.
        such as ['127.0.0.1:4150','182.168.1.1:4150']
    param: max_in_flight: number of messages get but not finish or req
    param: lookupd_http_addresses: first priority.if provided nsqd will neglected
    """
    loop = loop or asyncio.get_event_loop()
    if lookupd_http_addresses:
        reader = Reader(
            lookupd_http_addresses=lookupd_http_addresses,
            max_in_flight=max_in_flight,
            loop=loop,
        )
    else:
        if nsqd_tcp_addresses is None:
            nsqd_tcp_addresses = ["127.0.0.1:4150"]
        nsqd_tcp_addresses = [get_host_and_port(i) for i in nsqd_tcp_addresses]
        reader = Reader(
            nsqd_tcp_addresses=nsqd_tcp_addresses,
            max_in_flight=max_in_flight,
            loop=loop,
            **kwargs,
        )
    await reader.connect()
    return reader


class Reader:
    """
    NSQ tcp reader
    """

    def __init__(
        self,
        nsqd_tcp_addresses=None,
        lookupd_http_addresses=None,
        max_in_flight=42,
        loop=None,
        heartbeat_interval=30000,
        feature_negotiation=True,
        tls_v1=False,
        snappy=False,
        deflate=False,
        deflate_level=6,
        sample_rate=0,
        consumer=False,
        user_agent="",
    ):
        user_agents = ["nsqio/{}".format(get_version())]
        if user_agent:
            user_agents.append(user_agent)
        self._config = {
            "user_agent": " ".join(user_agents),
            "deflate": deflate,
            "deflate_level": deflate_level,
            "sample_rate": sample_rate,
            "snappy": snappy,
            "tls_v1": tls_v1,
            "heartbeat_interval": heartbeat_interval,
            "feature_negotiation": feature_negotiation,
        }
        self._nsqd_tcp_addresses = nsqd_tcp_addresses or []
        self._lookupd_http_addresses = lookupd_http_addresses or []

        self._max_in_flight = max_in_flight
        self._loop = loop or asyncio.get_event_loop()
        self._queue = asyncio.Queue(loop=self._loop)

        self._idle_timeout = 10

        self._max_in_flight = max_in_flight
        self._num_readers = 0

        self._is_subscribe = False
        self._redistribute_timeout = 5  # sec
        self._lookupd_poll_time = 30  # sec
        self.topic = None
        self.channel = None
        self._rdy_control = RdyControl(
            idle_timeout=self._idle_timeout,
            max_in_flight=self._max_in_flight,
            loop=self._loop,
        )
        self._auto_poll_lookupd_task_closed = asyncio.Event(loop=self._loop)
        self._auto_poll_lookupd_task = None

    async def connect(self):
        logging.info("reader connecting")
        if self._lookupd_http_addresses:
            """
            since lookupd must find a topic to find nsqds, 
            lookupd connect changed in to subscribe func
            """
            pass
        if self._nsqd_tcp_addresses:
            connections = {}
            for host, port in self._nsqd_tcp_addresses:
                conn: "TcpConnection" = await create_connection(
                    host, port, queue=self._queue, loop=self._loop
                )
                await self.prepare_conn(conn)
                connections[conn.id] = conn
            self._rdy_control.add_connections(connections)
        # init distribute for conns, init update rdy state for conn
        self._rdy_control.redistribute()

    async def prepare_conn(self, conn: "TcpConnection"):
        conn._on_message = partial(self._on_message, conn)
        _ = await conn.identify(**self._config)

    def _on_message(self, conn, msg):
        conn._last_message = time.time()
        if conn._on_rdy_changed_cb is not None:
            conn._on_rdy_changed_cb(conn.id)
        return msg

    async def _poll_lookupd(self, host, port):
        nsqlookup_conn = None
        try:
            nsqlookup_conn = NsqLookupd(host, port, loop=self._loop)

            res = await nsqlookup_conn.lookup(self.topic)
            logger.debug("lookupd response {}".format(res))

            if "producers" in res:
                addrs = []
                for producer in res["producers"]:
                    host = producer["broadcast_address"]
                    port = producer["tcp_port"]
                    addrs.append((host, port))
                return addrs
            else:
                logger.error("producers not found error")
                return []
        except Exception as e:
            logger.error(e)
            return []
        finally:
            if nsqlookup_conn is not None:
                try:
                    await nsqlookup_conn.close()
                except Exception as e:
                    logger.error(e)

    async def _init_lookupd_conns(self, host, port):
        producers = await self._poll_lookupd(host, port)
        if len(producers) == 0:
            # no producer
            return False
        for producer in producers:
            try:
                p_host, p_port = producer
                tmp_id = "tcp://{}:{}".format(p_host, p_port)
                if not await self._rdy_control._is_valid_connection(tmp_id):
                    logger.debug(
                        "new connection: host={}, port={}".format(p_host, p_port)
                    )
                    conn = await create_connection(
                        host=p_host, port=p_port, queue=self._queue, loop=self._loop
                    )
                    await self.prepare_conn(conn)
                    logger.debug("conn.id={}".format(conn.id))
                    # self._rdy_control.connections[conn.id] = conn
                    self._rdy_control.add_connection(conn)
            except Exception as e:
                logger.error(e)
                return False
        return True

    async def _auto_poll_lookupd(self):
        logger.debug("starting _auto_poll_lookupd")
        timeout_generator = retry_iterator(init_delay=1, max_delay=30.0)
        try:
            while (
                self._is_subscribe
                and self.topic is not None
                and self.channel is not None
            ):
                logger.info("reader _auto_poll_lookupd check loop")

                host, port = random.choice(self._lookupd_http_addresses)
                producers = await self._poll_lookupd(host, port)
                if len(producers) > 0:
                    # no producer
                    for producer in producers:
                        conn = None
                        try:
                            p_host, p_port = producer
                            tmp_id = "tcp://{}:{}".format(p_host, p_port)
                            # if tmp_id not in self._rdy_control.connections:
                            # missing? add it!
                            if not await self._rdy_control._is_valid_connection(tmp_id):
                                logger.debug(
                                    "_auto_poll_lookupd: new connection: host={}, port={}".format(
                                        p_host, p_port
                                    )
                                )
                                conn = await create_connection(
                                    host=p_host,
                                    port=p_port,
                                    queue=self._queue,
                                    loop=self._loop,
                                )
                                logger.debug("conn.id={}".format(conn.id))
                                await self.prepare_conn(conn)
                                # self._rdy_control.connections[conn.id] = conn
                                self._rdy_control.add_connection(conn)

                                await self.sub(conn, self.topic, self.channel)

                                logger.debug(
                                    "new nsqd added: conn.id={}".format(conn.id)
                                )

                                # why?
                                if conn._on_rdy_changed_cb is not None:
                                    conn._on_rdy_changed_cb(conn.id)
                            # else:
                            #     logger.debug("skip... {}".format(tmp_id))
                        except Exception as e:
                            logger.error(e)
                            if conn is not None:
                                try:
                                    conn.close()
                                except Exception as e:
                                    logger.error(e)
                        finally:
                            if not self._is_subscribe and conn is not None:
                                try:
                                    conn.close()
                                except Exception as e:
                                    logger.error(e)
                        if not self._is_subscribe:
                            break

                t = next(timeout_generator)
                await asyncio.sleep(t, loop=self._loop)
        except asyncio.CancelledError:
            logger.info("{} _auto_poll_lookupd".format(self))
        finally:
            self._auto_poll_lookupd_task_closed.set()

    async def subscribe(self, topic, channel):
        self.topic = topic
        self.channel = channel
        self._is_subscribe = True

        # firstly, we check lookupd
        if self._lookupd_http_addresses:
            lookupd_status = await self._lookupd()
            if not lookupd_status:
                logger.error("lookupd failed!")
                return False
            self._auto_poll_lookupd_task = self._loop.create_task(
                self._auto_poll_lookupd()
            )

        # and then, we sub all available topics
        for conn in self._rdy_control.connections.values():
            await self.sub(conn, topic, channel)
            if conn._on_rdy_changed_cb is not None:
                conn._on_rdy_changed_cb(conn.id)

        # redistribute is a task for fail or overload to
        # rebalance tcpconnections
        # consider for further. disable now
        # self._redistribute_task = self._loop.create_task(self._redistribute())
        return True

    async def sub(self, conn, topic, channel):
        await conn.execute(SUB, topic, channel)

    async def set_max_in_flight(self, max_in_flight):
        for conn in self._rdy_control.connections.values():
            await conn.execute(RDY, max_in_flight)

    async def send_cls(self):
        """ CLS 
            Cleanly close your connection (no more messages are sent)
        """
        for conn in self._rdy_control.connections.values():
            await conn.execute(CLS)

    async def messages(self):
        if not self._is_subscribe:
            logger.warning("You must subscribe to the topic first")
            return

        try:
            logger.debug("num readers ++ ")
            self._num_readers += 1
            while self._is_subscribe:
                result = await self._queue.get()
                if result is not None:
                    yield result
                else:
                    logger.debug("got NONE, skip")
            logger.debug("leaving message receiving")
        finally:
            logger.debug("num readers -- ")
            self._num_readers -= 1

    async def _redistribute(self):
        while self._is_subscribe:
            self._rdy_control.redistribute()
            await asyncio.sleep(self._redistribute_timeout, loop=self._loop)

    async def _lookupd(self):
        host, port = random.choice(self._lookupd_http_addresses)
        return await self._init_lookupd_conns(host, port)

    async def rescan_connections(self):
        return await self._lookupd()

    async def unsubscribe(self, timeout=10):
        if not self._is_subscribe:
            logger.warning("You must subscribe to the topic first")
            return
        # logger.debug("unsubscribing {}".format(self))
        # mark as disabled
        await self.set_max_in_flight(0)
        # clear is_subscribed flag
        self._is_subscribe = False

        # self._auto_poll_lookupd_task_closed = asyncio.Event(loop=self._loop)
        # self._auto_poll_lookupd_task = None
        try:
            if self._auto_poll_lookupd_task is not None:
                logger.info("canceling auto_poll_lookupd_task")
                self._auto_poll_lookupd_task.cancel()
                await asyncio.wait_for(
                    self._auto_poll_lookupd_task_closed.wait(),
                    timeout=timeout,
                    loop=self._loop,
                )
        except Exception as e:
            logger.error("cancel failed: {}".format(e))
        finally:
            logger.info("auto_poll_lookupd_task canceled")

        # send None to clear readers
        num_retries = 0
        while self._num_readers > 0 or num_retries > 1000:
            for i in range(self._num_readers):
                self._queue.put_nowait(None)
            num_retries += 1
            await asyncio.sleep(0.05)
        if self._num_readers > 0:
            logger.error(
                "{} retried {} times but still not work...".format(self, num_retries)
            )
        # no subscribers now

        # clear & req rest of the messages
        # mostly it is empty
        try:
            # consume all unfinished elems
            while not self._queue.empty():
                try:
                    result = self._queue.get_nowait()
                    if result is not None:
                        logger.debug("req: {}".format(result))
                        await result.req(0)
                except Exception as e:
                    logger.warning("req message failed {}".format(e))
        except Exception as e:
            logger.warning("requeue all messages failed {}".format(e))

    async def close(self):
        if self._is_subscribe:
            await self.unsubscribe()
        try:
            # await self.send_cls()
            # clear rdy_control
            if self._rdy_control is not None:
                self._rdy_control.stop_working()
            # close all connections
            # TODO: move to _rdy_control later
            for conn in self._rdy_control.connections.values():
                if conn is not None:
                    try:
                        conn.close()
                        await conn.wait_for_closed(1)
                    except Exception as e:
                        logger.error(e)
        except Exception as e:
            logger.error("close failed: {}".format(e))

    def __repr__(self):
        return "<Reader{}/{}>".format(self.topic, self.channel)



NL = b'\n'
DATA_SIZE = 4
FRAME_SIZE = 4
HEADER_SIZE = DATA_SIZE + FRAME_SIZE
MAX_CHUNK_SIZE = 65536

TIMESTAMP_SIZE = 8
ATTEMPTS_SIZE = 2
MSG_ID_SIZE = 16
MSG_HEADER = TIMESTAMP_SIZE + ATTEMPTS_SIZE + MSG_ID_SIZE
MAX_CHUNK_SIZE = 65

FRAME_TYPE_RESPONSE = 0
FRAME_TYPE_ERROR = 1
FRAME_TYPE_MESSAGE = 2
MAGIC_V2 = b'  V2'

HEARTBEAT = b'_heartbeat_'
PULSE = b'NOP\n'
BIN_OK = b'\x00\x00\x00\x06\x00\x00\x00\x00OK'

# nsq TCP commands
FIN = b'FIN'
REQ = b'REQ'
TOUCH = b'TOUCH'
RDY = b'RDY'
MPUB = b'MPUB'
CLS = b'CLS'
AUTH = b'AUTH'
SUB = b'SUB'
PUB = b'PUB'
DPUB = b'DPUB'

# connection status
CLOSED = 0
INIT = 1
CONNECTED = 2
SUBSCRIBED = 3
